/*
 * HD audio interface patch for Conexant HSF modems
 * 
 * Copyright (c) 2005-2006 Linuxant inc.
 * 
 * 1.  General Public License. This program is free software, and may
 * be redistributed or modified subject to the terms of the GNU General
 * Public License (version 2) or the GNU Lesser General Public License,
 * or (at your option) any later versions ("Open Source" code). You may
 * obtain a copy of the GNU General Public License at
 * http://www.fsf.org/copyleft/gpl.html and a copy of the GNU Lesser
 * General Public License at http://www.fsf.org/copyleft/less.html,
 * or you may alternatively write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA.
 * 
 * 2.   Disclaimer of Warranties. LINUXANT AND OTHER CONTRIBUTORS MAKE NO
 * REPRESENTATION ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.
 * IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTIES OF ANY KIND.
 * LINUXANT AND OTHER CONTRIBUTORS DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE, GOOD TITLE AND AGAINST INFRINGEMENT.
 * 
 * This software has not been formally tested, and there is no guarantee that
 * it is free of errors including, but not limited to, bugs, defects,
 * interrupted operation, or unexpected results. Any use of this software is
 * at user's own risk.
 * 
 * 3.   No Liability.
 * 
 * (a) Conexant or contributors shall not be responsible for any loss or
 * damage to Company, its customers, or any third parties for any reason
 * whatsoever, and LINUXANT OR CONTRIBUTORS SHALL NOT BE LIABLE FOR ANY
 * ACTUAL, DIRECT, INDIRECT, SPECIAL, PUNITIVE, INCIDENTAL, OR CONSEQUENTIAL
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED, WHETHER IN CONTRACT, STRICT OR OTHER LEGAL THEORY OF
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 * 
 * (b) User agrees to hold Conexant and contributors harmless from any
 * liability, loss, cost, damage or expense, including attorney's fees,
 * as a result of any claims which may be made by any person, including
 * but not limited to User, its agents and employees, its customers, or
 * any third parties that arise out of or result from the manufacture,
 * delivery, actual or alleged ownership, performance, use, operation
 * or possession of the software furnished hereunder, whether such claims
 * are based on negligence, breach of contract, absolute liability or any
 * other legal theory.
 * 
 * 4.   Notices. User hereby agrees not to remove, alter or destroy any
 * copyright, trademark, credits, other proprietary notices or confidential
 * legends placed upon, contained within or associated with the Software,
 * and shall include all such unaltered copyright, trademark, credits,
 * other proprietary notices or confidential legends on or in every copy of
 * the Software.
 * 
 */
#include <sound/driver.h>
#include <linux/init.h>
#include <sound/core.h>
#include "hda_codec.h"
#include "hda_local.h"

#if defined(__i386__)
#define __shimcall__ __attribute__((regparm(0)))
#define __kernelcall__
#define __kernelcallregparm__ __attribute__((regparm(3)))
#define __kernelcallstkparm__ __attribute__((regparm(0)))
#else
#define __shimcall__
#define __kernelcall__
#define __kernelcallregparm__
#define __kernelcallstkparm__
#endif

struct cnxthsf_spec {
	void *pDevNode;

	unsigned int stream_tags[2];
	int do_prepare[2];

	void (*cbHdaEvent)(void *Context, unsigned int res) __shimcall__;
	void *cbHdaEventContext;

	unsigned char cbHdaTag;

	struct hda_pcm pcm;
};

static int cnxthsf_pcm_prepare(struct hda_pcm_stream *hinfo,
			      struct hda_codec *codec,
			      unsigned int stream_tag,
			      unsigned int format,
			      struct snd_pcm_substream *substream)
{
	struct cnxthsf_spec *spec = codec->spec;

//printk(KERN_DEBUG"%s: codec=%p stream=%d stream_tag=%x format=0x%x substream=%p\n", __FUNCTION__, codec, substream->stream, stream_tag, format, substream);

	spec->stream_tags[substream->stream] = stream_tag;

	return 0;
}

static int cnxthsf_pcm_open(struct hda_pcm_stream *hinfo, struct hda_codec *codec,
			    struct snd_pcm_substream *substream)
{
	static unsigned int rates[] = { 16000 };
	static struct snd_pcm_hw_constraint_list hw_constraints_rates = {
		.count = ARRAY_SIZE(rates),
		.list = rates,
		.mask = 0,
	};

//printk(KERN_DEBUG"%s: codec=%p substream=%p\n", __FUNCTION__, codec, substream);

#if 0 // XXX
	//substream->runtime->hw.periods_min = 1;
	substream->runtime->hw.period_bytes_min = 256; //4096;
#endif
	return snd_pcm_hw_constraint_list(substream->runtime, 0,
			SNDRV_PCM_HW_PARAM_RATE, &hw_constraints_rates);
}

#if 0
static int cnxthsf_pcm_close(struct hda_pcm_stream *info, struct hda_codec *codec, struct snd_pcm_substream *substream)
{
}

static int cnxthsf_pcm_cleanup(struct hda_pcm_stream *info, struct hda_codec *codec, struct snd_pcm_substream *substream)
{
}
#endif

static struct hda_pcm_stream cnxthsf_pcm = {
	.substreams = 1,
	.channels_min = 1,
	.channels_max = 1,
	.nid = 0x1,
	.rates = SNDRV_PCM_RATE_16000,
	.formats = SNDRV_PCM_FMTBIT_S16_LE,
	.maxbps = 16,
	.ops = {
		.open = cnxthsf_pcm_open,
		//.close = cnxthsf_pcm_close,
		.prepare = cnxthsf_pcm_prepare,
		//.cleanup = cnxthsf_pcm_cleanup,
	},
};



static int cnxthsf_build_pcms(struct hda_codec *codec)
{
	struct cnxthsf_spec *spec = codec->spec;
	struct hda_pcm *info = &spec->pcm;

//printk(KERN_DEBUG"%s: codec=%p mfg=%d\n", __FUNCTION__, codec, codec->mfg);

	cnxthsf_pcm.nid = codec->mfg;
	codec->num_pcms = 1;
	codec->pcm_info = info;
	info->name = "HSF Modem";
	info->stream[SNDRV_PCM_STREAM_PLAYBACK] = cnxthsf_pcm;
	info->stream[SNDRV_PCM_STREAM_CAPTURE]  = cnxthsf_pcm;
	info->is_modem = 1;
	return 0;
}


static int cnxthsf_init(struct hda_codec *codec)
{
	int ret;
	struct cnxthsf_spec *spec = codec->spec;
	int (*cnxthwhda_probe)(void *codec, struct device *hwDev, void **ppDevNode);

//printk(KERN_DEBUG"%s: codec=%p\n", __FUNCTION__, codec);
    snd_hda_codec_write(codec, AC_NODE_ROOT, 0, AC_VERB_SET_CODEC_RESET, 0);

	cnxthwhda_probe = (void*)symbol_request(cnxthwhda_probe);
	if(!cnxthwhda_probe) {
		printk(KERN_ERR"%s: symbol_request(cnxthwhda_probe) failed\n", __FUNCTION__);
		return -ENOENT;
	}

	ret = cnxthwhda_probe(codec, codec->bus->card->dev, &spec->pDevNode);
	if(ret) {
		printk(KERN_ERR"%s: cnxthwhda_probe() failed: %d\n", __FUNCTION__, ret);
		symbol_put(cnxthwhda_probe);
		return ret;
	}

	return ret;
}

static int cnxthsf_exit(struct hda_codec *codec)
{
	struct cnxthsf_spec *spec = codec->spec;

//printk(KERN_DEBUG"%s: codec=%p spec=%p\n", __FUNCTION__, codec, spec);

	if(spec && spec->pDevNode) {
		void (*cnxthwhda_remove)(void *ptr);

		cnxthwhda_remove = (void*)symbol_request(cnxthwhda_remove);
		if(cnxthwhda_remove) {
			cnxthwhda_remove(spec->pDevNode);
			spec->pDevNode = NULL;
			symbol_put(cnxthwhda_remove);
			symbol_put(cnxthwhda_probe);
		} else {
			printk(KERN_ERR"%s: symbol_request(cnxthwhda_remove) failed\n", __FUNCTION__);
		}
	}

	return 0;
}


static void cnxthsf_free(struct hda_codec *codec)
{
	struct cnxthsf_spec *spec = codec->spec;

//printk(KERN_DEBUG"%s: codec=%p spec=%p\n", __FUNCTION__, codec, spec);

	if(spec) {
		codec->spec = NULL;
		memset(&codec->patch_ops, 0, sizeof(codec->patch_ops));
		kfree(spec);
	}
}


static void cnxthsf_unsol_event(struct hda_codec *codec, unsigned int res)
{
	struct cnxthsf_spec *spec = codec->spec;

	//printk(KERN_DEBUG"%s: codec=%p res=0x%02x spec=%p cbHdaEvent=%p\n", __FUNCTION__, codec, res, spec, spec->cbHdaEvent);

	if(spec && spec->cbHdaEvent) {
		if(((res >> 26) & 0x3f) == spec->cbHdaTag) {
			//printk(KERN_DEBUG"%s: res=0x%02x cbHdaEvent=%p\n", __FUNCTION__, res, spec->cbHdaEvent);
//			printk(KERN_DEBUG"%s: calling cbHdaEvent=%p ctx=%p\n", __FUNCTION__, spec->cbHdaEvent, spec->cbHdaEventContext);
			spec->cbHdaEvent(spec->cbHdaEventContext, res);
		} else {
			printk(KERN_DEBUG"%s: ignoring res=0x08%x\n", __FUNCTION__, res);
		}
	}
}

static int cnxthsf_suspend(struct hda_codec *codec, pm_message_t state)
{
	struct cnxthsf_spec *spec = codec->spec;
	int (*cnxthwhda_suspend)(void *devnode, pm_message_t state);
	int ret;

	//printk(KERN_DEBUG"%s: codec=%p spec=%p state=0x%x\n", __FUNCTION__, codec, spec, *((u32 *)&state));

	if(!spec || !spec->pDevNode) {
		return -EINVAL;
	}

	cnxthwhda_suspend = (void*)symbol_request(cnxthwhda_suspend);
	if(!cnxthwhda_suspend) {
		printk(KERN_ERR"%s: symbol_request(cnxthwhda_suspend) failed\n", __FUNCTION__);
		return -ENOSYS;
	}

	ret = cnxthwhda_suspend(spec->pDevNode, state);
	symbol_put(cnxthwhda_suspend);

	return ret;
}

static int cnxthsf_resume (struct hda_codec *codec)
{
	struct cnxthsf_spec *spec = codec->spec;
	int (*cnxthwhda_resume)(void *devnode);
	int ret;

	//printk(KERN_DEBUG"%s: codec=%p spec=%p\n", __FUNCTION__, codec, spec);

	if(!spec || !spec->pDevNode) {
		return -EINVAL;
	}

	cnxthwhda_resume = (void*)symbol_request(cnxthwhda_resume);
	if(!cnxthwhda_resume) {
		printk(KERN_ERR"%s: symbol_request(cnxthwhda_resume) failed\n", __FUNCTION__);
		return -ENOSYS;
	}

    snd_hda_codec_write(codec, AC_NODE_ROOT, 0, AC_VERB_SET_CODEC_RESET, 0);

	ret = cnxthwhda_resume(spec->pDevNode);
	symbol_put(cnxthwhda_resume);

	return ret;
}


static struct hda_codec_ops cnxthsf_patch_ops = {
	.build_pcms = cnxthsf_build_pcms,
	.init = cnxthsf_init,
	.exit = cnxthsf_exit,
	.free = cnxthsf_free,
	.unsol_event = cnxthsf_unsol_event,
#ifdef CONFIG_PM
	.suspend = cnxthsf_suspend,
	.resume = cnxthsf_resume,
#endif
};
      

static int patch_cnxthsf(struct hda_codec *codec)
{
	struct cnxthsf_spec *spec;
	int (*cnxthwhda_probe)(void *codec, struct device *hwDev, void **ppDevNode);

//printk(KERN_DEBUG"%s: codec=%p\n", __FUNCTION__, codec);

	if(!codec->mfg) { // we only support modems
		return -ENODEV;
	}

	spec = kcalloc(1, sizeof(*spec), GFP_KERNEL);
	if (spec == NULL)
		return -ENOMEM;
	codec->spec = spec;
	codec->patch_ops = cnxthsf_patch_ops;

	cnxthwhda_probe = (void*)symbol_request(cnxthwhda_probe);
	if(cnxthwhda_probe) {
		symbol_put(cnxthwhda_probe);
	} else {
		printk(KERN_ERR"%s: symbol_request(cnxthwhda_probe) failed\n", __FUNCTION__);
		codec->spec = NULL;
		memset(&codec->patch_ops, 0, sizeof(codec->patch_ops));
		kfree(spec);
		return -ENOENT;
	}

	return 0;
}


struct hda_codec_preset snd_hda_preset_cnxthsf[] = {
 	{ .id = 0x14f10000, .mask = 0xffff0000, .mfg = 2, .name = "HSF", .patch = patch_cnxthsf },
	{}
};




typedef unsigned long   ULONG,  *PULONG;

#include "../../include/oshda.h"

__shimcall__
unsigned int OsHdaCodecGetAddr(PHDAOSHAL pHdaOsHal)
{
	return ((struct hda_codec *)pHdaOsHal->hda_codec)->addr;
}
EXPORT_SYMBOL(OsHdaCodecGetAddr);

__shimcall__
unsigned int OsHdaCodecGetVendorId(PHDAOSHAL pHdaOsHal)
{
	return ((struct hda_codec *)pHdaOsHal->hda_codec)->vendor_id;
}
EXPORT_SYMBOL(OsHdaCodecGetVendorId);

__shimcall__
unsigned int OsHdaCodecGetSubsystemId(PHDAOSHAL pHdaOsHal)
{
	return ((struct hda_codec *)pHdaOsHal->hda_codec)->subsystem_id;
}
EXPORT_SYMBOL(OsHdaCodecGetSubsystemId);

__shimcall__
unsigned int OsHdaCodecGetRevisionId(PHDAOSHAL pHdaOsHal)
{
	return ((struct hda_codec *)pHdaOsHal->hda_codec)->revision_id;
}
EXPORT_SYMBOL(OsHdaCodecGetRevisionId);

__shimcall__
unsigned int OsHdaCodecRead(PHDAOSHAL pHdaOsHal, unsigned short nid, int direct, unsigned int verb, unsigned int para)
{
#if 1
	return snd_hda_codec_read((struct hda_codec *)pHdaOsHal->hda_codec, nid, direct, verb, para);
#else
	unsigned int res;

	res = snd_hda_codec_read((struct hda_codec *)pHdaOsHal->hda_codec, nid, direct, verb, para);

	printk(KERN_DEBUG"%s: nid=%x direct=%d verb=0x%x para=0x%x res=0x%08x\n", __FUNCTION__, nid, direct, verb, para, res);

	return res;
#endif
}
EXPORT_SYMBOL(OsHdaCodecRead);

__shimcall__
unsigned int OsHdaCodecWallclock(PHDAOSHAL pHdaOsHal)
{
	return snd_hda_codec_wallclock((struct hda_codec *)pHdaOsHal->hda_codec);
}
EXPORT_SYMBOL(OsHdaCodecWallclock);

__shimcall__
void OsHdaCodecSetEventCallback(PHDAOSHAL pHdaOsHal, void (*cbHdaEvent)(void *Context, unsigned int res), void *cbHdaEventContext, unsigned char *cbHdaTag)
{
	struct cnxthsf_spec *spec = ((struct hda_codec *)pHdaOsHal->hda_codec)->spec;

	*cbHdaTag = ((struct hda_codec *)pHdaOsHal->hda_codec)->mfg;

//	printk(KERN_DEBUG"%s: codec=%p %p(%p) tag=%d\n", __FUNCTION__, pHdaOsHal->hda_codec, cbHdaEvent, cbHdaEventContext, (int)*cbHdaTag);

	spec->cbHdaTag = *cbHdaTag;
	spec->cbHdaEventContext = cbHdaEventContext;
	spec->cbHdaEvent = (void*)cbHdaEvent;
}
EXPORT_SYMBOL(OsHdaCodecSetEventCallback);

__shimcall__
void OsHdaCodecClearEventCallback(PHDAOSHAL pHdaOsHal, unsigned char cbHdaTag)
{
	struct cnxthsf_spec *spec = ((struct hda_codec *)pHdaOsHal->hda_codec)->spec;

	if(spec) {
//		printk(KERN_DEBUG"%s: codec=%p Tag=%d\n", __FUNCTION__, pHdaOsHal->hda_codec, cbHdaTag);
		spec->cbHdaEvent = NULL;
		spec->cbHdaEventContext = NULL;
		spec->cbHdaTag = 0;
	}
}
EXPORT_SYMBOL(OsHdaCodecClearEventCallback);


#include <sound/pcm_params.h>

#ifndef FOUND_SND_PCM_HW_PARAM_SET
static int snd_interval_refine_set(struct snd_interval *i, unsigned int val)
{
	struct snd_interval t;
	t.empty = 0;
	t.min = t.max = val;
	t.openmin = t.openmax = 0;
	t.integer = 1;
	return snd_interval_refine(i, &t);
}

static int _snd_pcm_hw_param_set(struct snd_pcm_hw_params *params,
				 snd_pcm_hw_param_t var, unsigned int val,
				 int dir)
{
	int changed;
	if (hw_is_mask(var)) {
		struct snd_mask *m = hw_param_mask(params, var);
		if (val == 0 && dir < 0) {
			changed = -EINVAL;
			snd_mask_none(m);
		} else {
			if (dir > 0)
				val++;
			else if (dir < 0)
				val--;
			changed = snd_mask_refine_set(hw_param_mask(params, var), val);
		}
	} else if (hw_is_interval(var)) {
		struct snd_interval *i = hw_param_interval(params, var);
		if (val == 0 && dir < 0) {
			changed = -EINVAL;
			snd_interval_none(i);
		} else if (dir == 0)
			changed = snd_interval_refine_set(i, val);
		else {
			struct snd_interval t;
			t.openmin = 1;
			t.openmax = 1;
			t.empty = 0;
			t.integer = 0;
			if (dir < 0) {
				t.min = val - 1;
				t.max = val;
			} else {
				t.min = val;
				t.max = val+1;
			}
			changed = snd_interval_refine(i, &t);
		}
	} else
		return -EINVAL;
	if (changed) {
		params->cmask |= 1 << var;
		params->rmask |= 1 << var;
	}
	return changed;
}
#endif /* !FOUND_SND_PCM_HW_PARAM_SET */

static int cnxthsf_snd_pcm_change_params(struct snd_pcm_substream *substream, int hw_param_buffer_bytes)
{
	//snd_pcm_runtime_t *runtime = substream->runtime;
	snd_pcm_hw_params_t *sparams;
	int err;

	sparams = kmalloc(sizeof(*sparams), GFP_KERNEL);
	if (/*!params ||*/ !sparams) {
		return -ENOMEM;
	}

	_snd_pcm_hw_params_any(sparams);
	_snd_pcm_hw_param_set(sparams, SNDRV_PCM_HW_PARAM_BUFFER_BYTES,
                                      hw_param_buffer_bytes, 0);

#if 0
	_snd_pcm_hw_param_set(sparams, SNDRV_PCM_HW_PARAM_PERIOD_SIZE,
                                      hw_param_buffer_bytes / 2, 0);
#endif

	snd_pcm_kernel_ioctl(substream, SNDRV_PCM_IOCTL_DROP, NULL);

	if ((err = snd_pcm_kernel_ioctl(substream, SNDRV_PCM_IOCTL_HW_PARAMS, sparams)) < 0) {
		printk(KERN_ERR"%s: SNDRV_PCM_IOCTL_HW_PARAMS failed (%d)\n", __FUNCTION__, err);
		//return err;
	}

	return 0;
}

static int cnxthsf_snd_pcm_prepare_substream(PHDAOSHAL pHdaOsHal, struct snd_pcm_substream *substream)
{
	int err;
	struct cnxthsf_spec *spec = ((struct hda_codec *)pHdaOsHal->hda_codec)->spec;

//printk(KERN_DEBUG"%s: substream=%p\n", __FUNCTION__, substream);

	if(spec->do_prepare[substream->stream]) {
		err = snd_pcm_kernel_ioctl(substream, SNDRV_PCM_IOCTL_PREPARE, NULL);
		if (err < 0) {
			printk(KERN_ERR"%s: SNDRV_PCM_IOCTL_PREPARE failed (%d)\n", __FUNCTION__, err);
			return err;
		}
		spec->do_prepare[substream->stream] = 0;
	}
#if 0
	runtime->oss.prepare = 0;
	runtime->oss.prev_hw_ptr_interrupt = 0;
	runtime->oss.period_ptr = 0;
	runtime->oss.buffer_used = 0;
#endif

	return 0;
}

__shimcall__
int OsHdaCodecOpenDMA(PHDAOSHAL pHdaOsHal, int hw_param_buffer_bytes, void **ppPlaybackStream, void **ppCaptureStream)
{
	int err;
	snd_pcm_t *pcm = ((struct hda_codec *)pHdaOsHal->hda_codec)->pcm_info->pcm;
	struct cnxthsf_spec *spec = ((struct hda_codec *)pHdaOsHal->hda_codec)->spec;

	struct snd_pcm_substream *psubstream = NULL, *csubstream = NULL;
#ifdef FOUND_OPEN_SUBSTREAM_FILE
	static struct file fil;
#endif

//printk(KERN_DEBUG"%s: pcm=%p hw_param_buffer_bytes=%d\n", __FUNCTION__, pcm, hw_param_buffer_bytes);

#ifdef FOUND_OPEN_SUBSTREAM_FILE
	if ((err = snd_pcm_open_substream(pcm, SNDRV_PCM_STREAM_PLAYBACK, &fil,
#else
	if ((err = snd_pcm_open_substream(pcm, SNDRV_PCM_STREAM_PLAYBACK,
#endif
					&psubstream)) < 0) {
		printk(KERN_ERR"%s: snd_pcm_open_substream STREAM_PLAYBACK failed (%d)\n", __FUNCTION__, err);
		return err;
	}

#ifdef FOUND_OPEN_SUBSTREAM_FILE
	if ((err = snd_pcm_open_substream(pcm, SNDRV_PCM_STREAM_CAPTURE, &fil,
#else
	if ((err = snd_pcm_open_substream(pcm, SNDRV_PCM_STREAM_CAPTURE,
#endif
					&csubstream)) < 0) {
		printk(KERN_ERR"%s: snd_pcm_open_substream STREAM_CAPTURE failed (%d)\n", __FUNCTION__, err);
		return err;
	}

#ifndef FOUND_OPEN_SUBSTREAM_FILE
	err = snd_pcm_hw_constraints_init(psubstream);
	if (err < 0) {
		printk(KERN_ERR"%s: snd_pcm_hw_constraints_init STREAM_PLAYBACK failed (%d)\n", __FUNCTION__, err);
	}
	if ((err = psubstream->ops->open(psubstream)) < 0) {
		printk(KERN_ERR"%s: ops open STREAM_PLAYBACK failed (%d)\n", __FUNCTION__, err);
	}
	//psubstream->ffile = file;
	err = snd_pcm_hw_constraints_complete(psubstream);
	if (err < 0) {
		printk(KERN_ERR"%s: snd_pcm_hw_constraints_complete STREAM_PLAYBACK failed (%d)\n", __FUNCTION__, err);
	}

	err = snd_pcm_hw_constraints_init(csubstream);
	if (err < 0) {
		printk(KERN_ERR"%s: snd_pcm_hw_constraints_init STREAM_CAPTURE failed (%d)\n", __FUNCTION__, err);
	}
	if ((err = csubstream->ops->open(csubstream)) < 0) {
		printk(KERN_ERR"%s: ops open STREAM_CAPTURE failed (%d)\n", __FUNCTION__, err);
	}
	//csubstream->ffile = file;
	err = snd_pcm_hw_constraints_complete(csubstream);
	if (err < 0) {
		printk(KERN_ERR"%s: snd_pcm_hw_constraints_complete STREAM_CAPTURE failed (%d)\n", __FUNCTION__, err);
	}
#endif

	err = cnxthsf_snd_pcm_change_params(psubstream, hw_param_buffer_bytes);
	if (err < 0) {
		printk(KERN_ERR"%s: cnxthsf_snd_pcm_change_params STREAM_PLAYBACK failed (%d)\n", __FUNCTION__, err);
	}

	err = cnxthsf_snd_pcm_change_params(csubstream, hw_param_buffer_bytes);
	if (err < 0) {
		printk(KERN_ERR"%s: cnxthsf_snd_pcm_change_params STREAM_CAPTURE failed (%d)\n", __FUNCTION__, err);
	}

#if 0
	printk(KERN_DEBUG"%s: psubstream=%p dma_buffer_p=%p area=%p addr=0x%lx bytes=%d\n", __FUNCTION__,
			psubstream,
			psubstream->runtime->dma_buffer_p,
			psubstream->runtime->dma_area,
			(unsigned long)psubstream->runtime->dma_addr,
			psubstream->runtime->dma_bytes);

	printk(KERN_DEBUG"%s: csubstream=%p dma_buffer_p=%p area=%p addr=0x%lx bytes=%d\n", __FUNCTION__,
			csubstream,
			csubstream->runtime->dma_buffer_p,
			csubstream->runtime->dma_area,
			(unsigned long)csubstream->runtime->dma_addr,
			csubstream->runtime->dma_bytes);
#endif

#if 0
	// Link both streams together
	if (!snd_pcm_stream_linked(psubstream)) {
		psubstream->group = kmalloc(sizeof(snd_pcm_group_t), GFP_ATOMIC);
		if (psubstream->group == NULL) {
			res = -ENOMEM;
			goto _end;
		}
		spin_lock_init(&psubstream->group->lock);
		INIT_LIST_HEAD(&psubstream->group->psubstreams);
		list_add_tail(&psubstream->link_list, &psubstream->group->substreams);
		psubstream->group->count = 1;
	}
	list_add_tail(&csubstream->link_list, &psubstream->group->substreams);
	psubstream->group->count++;
	csubstream->group = psubstream->group;
#endif


	spec->do_prepare[psubstream->stream] = 1;
	spec->do_prepare[csubstream->stream] = 1;

	if ((err = cnxthsf_snd_pcm_prepare_substream(pHdaOsHal, psubstream)) < 0)
		return err;

	if ((err = cnxthsf_snd_pcm_prepare_substream(pHdaOsHal, csubstream)) < 0)
		return err;

	*ppPlaybackStream = psubstream;
	*ppCaptureStream = csubstream;

	return 0;
}
EXPORT_SYMBOL(OsHdaCodecOpenDMA);

static void cnxthsf_snd_pcm_close_stream(struct snd_pcm_substream *substream)
{
//	printk(KERN_DEBUG"%s: substream=%p\n", __FUNCTION__, substream);

	snd_pcm_stream_lock_irq(substream);
	if (snd_pcm_running(substream))
		snd_pcm_stop(substream, SNDRV_PCM_STATE_SETUP);
	snd_pcm_stream_unlock_irq(substream);

	if (substream->ops->hw_free != NULL)
		substream->ops->hw_free(substream);

	substream->ops->close(substream);
	snd_pcm_release_substream(substream);
}

__shimcall__
void OsHdaCodecCloseDMA(PHDAOSHAL pHdaOsHal, void *pPlaybackStream, void *pCaptureStream)
{
	if(pHdaOsHal && pHdaOsHal->bActive) {
		if(pPlaybackStream)
			cnxthsf_snd_pcm_close_stream(pPlaybackStream);
		if(pCaptureStream)
			cnxthsf_snd_pcm_close_stream(pCaptureStream);
	}
}
EXPORT_SYMBOL(OsHdaCodecCloseDMA);

__shimcall__
void OsHdaCodecDMAInfo(PHDAOSHAL pHdaOsHal, void *Stream, unsigned char *StreamID, ULONG *FifoSize, short unsigned int **ppBufAddr)
{
	struct cnxthsf_spec *spec = ((struct hda_codec *)pHdaOsHal->hda_codec)->spec;
	struct snd_pcm_substream *substream = (struct snd_pcm_substream *)Stream;

#if 0
	printk(KERN_DEBUG"%s: substream=%p stream=%u id/tag=%u fifo_size=%u bufAddr=%p\n", __FUNCTION__,
			substream,
			substream->stream,
			spec->stream_tags[substream->stream],
			substream->runtime->hw.fifo_size,
			substream->runtime->dma_area);
#endif

	*StreamID = spec->stream_tags[substream->stream];
	*FifoSize = substream->runtime->hw.fifo_size; // XXX
	*ppBufAddr = (short unsigned int *)substream->runtime->dma_area;
}
EXPORT_SYMBOL(OsHdaCodecDMAInfo);

__shimcall__
int OsHdaCodecSetDMAState(PHDAOSHAL pHdaOsHal, OSHDA_STREAM_STATE streamState, void *pPlaybackStream, void *pCaptureStream)
{
	struct cnxthsf_spec *spec = ((struct hda_codec *)pHdaOsHal->hda_codec)->spec;
        struct snd_pcm_substream *psubstream = (struct snd_pcm_substream *)pPlaybackStream;
        struct snd_pcm_substream *csubstream = (struct snd_pcm_substream *)pCaptureStream;
	int err = 0, cmd;
	unsigned long flags, flags2;

	switch(streamState) {
		case OsHdaStreamStateRun:
//			printk(KERN_DEBUG"%s: Run\n", __FUNCTION__);
			if ((err = cnxthsf_snd_pcm_prepare_substream(pHdaOsHal, psubstream)) < 0)
				return err;
			if ((err = cnxthsf_snd_pcm_prepare_substream(pHdaOsHal, csubstream)) < 0)
				return err;

			cmd = SNDRV_PCM_IOCTL_START;
			psubstream->runtime->start_threshold = 1;
			psubstream->runtime->stop_threshold = psubstream->runtime->boundary;
			csubstream->runtime->start_threshold = 1;
			csubstream->runtime->stop_threshold = csubstream->runtime->boundary;
			break;
		case OsHdaStreamStateStop:
//			printk(KERN_DEBUG"%s: Stop\n", __FUNCTION__);
			cmd = SNDRV_PCM_IOCTL_DROP;
			break;
		case OsHdaStreamStateReset:
//			printk(KERN_DEBUG"%s: Reset\n", __FUNCTION__);
			cmd = SNDRV_PCM_IOCTL_RESET;
			psubstream->runtime->start_threshold = psubstream->runtime->boundary;
			csubstream->runtime->start_threshold = csubstream->runtime->boundary;
			break;
		default:
			printk(KERN_ERR"%s: unknown state %d\n", __FUNCTION__, streamState);
			return -ENOSYS;
	}

	switch(cmd) {
		case SNDRV_PCM_IOCTL_START:
			snd_pcm_stream_lock_irqsave(psubstream, flags);
			snd_pcm_stream_lock_irqsave(csubstream, flags2);
			psubstream->ops->trigger(psubstream, SNDRV_PCM_TRIGGER_START);
			csubstream->ops->trigger(csubstream, SNDRV_PCM_TRIGGER_START);
			snd_pcm_stream_unlock_irqrestore(csubstream, flags2);
			snd_pcm_stream_unlock_irqrestore(psubstream, flags);
			break;
		case SNDRV_PCM_IOCTL_DROP:
			snd_pcm_stream_lock_irqsave(psubstream, flags);
			snd_pcm_stream_lock_irqsave(csubstream, flags2);
			psubstream->runtime->start_threshold = psubstream->runtime->boundary;
			csubstream->runtime->start_threshold = csubstream->runtime->boundary;
			psubstream->ops->trigger(psubstream, SNDRV_PCM_TRIGGER_STOP);
			csubstream->ops->trigger(csubstream, SNDRV_PCM_TRIGGER_STOP);
			snd_pcm_stream_unlock_irqrestore(csubstream, flags2);
			snd_pcm_stream_unlock_irqrestore(psubstream, flags);
			break;
		case SNDRV_PCM_IOCTL_RESET:
			//psubstream->ops->trigger(psubstream, SNDRV_PCM_TRIGGER_STOP);
			//csubstream->ops->trigger(csubstream, SNDRV_PCM_TRIGGER_STOP);
			/*FALLTHROUGH*/
		default:
#ifdef FOUND_OPEN_SUBSTREAM_FILE
			err = snd_pcm_kernel_ioctl(psubstream, cmd, NULL);
#else
			err = snd_pcm_kernel_playback_ioctl(psubstream, cmd, NULL);
#endif
			if (err < 0) {
				printk(KERN_ERR"%s: snd_pcm_kernel_ioctl (playback) failed (%d)\n", __FUNCTION__, err);
			}

#ifdef FOUND_OPEN_SUBSTREAM_FILE
			err = snd_pcm_kernel_ioctl(csubstream, cmd, NULL);
#else
			err = snd_pcm_kernel_capture_ioctl(csubstream, cmd, NULL);
#endif
			if (err < 0) {
				printk(KERN_ERR"%s: snd_pcm_kernel_ioctl (capture) failed (%d)\n", __FUNCTION__, err);
			}
			break;
	}

	if(cmd != SNDRV_PCM_IOCTL_START) {
		spec->do_prepare[psubstream->stream] = 1;
		spec->do_prepare[csubstream->stream] = 1;
	}

	return err;
}
EXPORT_SYMBOL(OsHdaCodecSetDMAState);

__shimcall__
unsigned long OsHdaCodecGetDMAPos(PHDAOSHAL pHdaOsHal, void *Stream)
{
	struct hda_codec *codec = (struct hda_codec *)pHdaOsHal->hda_codec;
	struct snd_pcm_substream *substream = (struct snd_pcm_substream *)Stream;
	int ret;

        if(codec->bus->ops.get_linkpos)
                ret = codec->bus->ops.get_linkpos(substream);
	else
		ret = frames_to_bytes(substream->runtime, substream->ops->pointer(substream));

	//printk(KERN_DEBUG"%s: substream=%p pos=%ld, ret=%d\n", __FUNCTION__, substream, pos, ret);

	return ret;
}
EXPORT_SYMBOL(OsHdaCodecGetDMAPos);

